@@ -35,7 +35,18 @@
  * soup_socket_get_remote_address()) may be useful to applications.
  **/
 
-G_DEFINE_TYPE (SoupSocket, soup_socket, G_TYPE_OBJECT)
+static void
+_soup_networking_init (void)
+{
+#ifdef G_OS_WIN32
+  WSADATA wsadata;
+  if (WSAStartup (MAKEWORD (2, 0), &wsadata) != 0)
+    g_error ("Windows Sockets could not be initialized");
+#endif
+}
+
+G_DEFINE_TYPE_WITH_CODE (SoupSocket, soup_socket, G_TYPE_OBJECT,
+			 _soup_networking_init ();)
 
 enum {
 	READABLE,
@@ -55,8 +66,10 @@ enum {
 	PROP_NON_BLOCKING,
 	PROP_IS_SERVER,
 	PROP_SSL_CREDENTIALS,
+	PROP_SSL_STRICT,
 	PROP_ASYNC_CONTEXT,
 	PROP_TIMEOUT,
+	PROP_TRUSTED_CERTIFICATE,
 
 	LAST_PROP
 };
@@ -70,6 +83,8 @@ typedef struct {
 	guint is_server:1;
 	guint timed_out:1;
 	gpointer ssl_creds;
+	gboolean ssl_strict;
+	gboolean trusted_certificate;
 
 	GMainContext   *async_context;
 	GSource        *watch_src;
@@ -342,6 +357,33 @@ soup_socket_class_init (SoupSocketClass *socket_class)
 				      "SSL credentials",
 				      "SSL credential information, passed from the session to the SSL implementation",
 				      G_PARAM_READWRITE));
+	/**
+	 * SOUP_SOCKET_SSL_STRICT:
+	 *
+	 * Alias for the #SoupSocket:ignore-ssl-cert-errors property.
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_SSL_STRICT,
+		g_param_spec_boolean (SOUP_SOCKET_SSL_STRICT,
+				      "Strictly validate SSL certificates",
+				      "Whether certificate errors should be considered a connection error",
+				      TRUE,
+				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_SOCKET_TRUSTED_CERTIFICATE:
+	 *
+	 * Alias for the #SoupSocket:trusted-certificate
+	 * property. Notice that this property's value is only useful
+	 * if the socket is for an SSL connection, and only reliable
+	 * after some data has been transferred to or from it.
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_TRUSTED_CERTIFICATE,
+		g_param_spec_boolean (SOUP_SOCKET_TRUSTED_CERTIFICATE,
+				     "Trusted Certificate",
+				     "Whether the server certificate is trusted, if this is an SSL socket",
+				     FALSE,
+				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 	/**
 	 * SOUP_SOCKET_ASYNC_CONTEXT:
 	 *
@@ -368,11 +410,6 @@ soup_socket_class_init (SoupSocketClass *socket_class)
 				   "Value in seconds to timeout a blocking I/O",
 				   0, G_MAXUINT, 0,
 				   G_PARAM_READWRITE));
-
-#ifdef G_OS_WIN32
-	/* Make sure WSAStartup() gets called. */
-	soup_address_get_type ();
-#endif
 }
 
 
@@ -486,6 +523,12 @@ set_property (GObject *object, guint prop_id,
 	case PROP_SSL_CREDENTIALS:
 		priv->ssl_creds = g_value_get_pointer (value);
 		break;
+	case PROP_SSL_STRICT:
+		priv->ssl_strict = g_value_get_boolean (value);
+		break;
+	case PROP_TRUSTED_CERTIFICATE:
+		priv->trusted_certificate = g_value_get_boolean (value);
+		break;
 	case PROP_ASYNC_CONTEXT:
 		priv->async_context = g_value_get_pointer (value);
 		if (priv->async_context)
@@ -522,6 +565,12 @@ get_property (GObject *object, guint prop_id,
 	case PROP_SSL_CREDENTIALS:
 		g_value_set_pointer (value, priv->ssl_creds);
 		break;
+	case PROP_SSL_STRICT:
+		g_value_set_boolean (value, priv->ssl_strict);
+		break;
+	case PROP_TRUSTED_CERTIFICATE:
+		g_value_set_boolean (value, priv->trusted_certificate);
+		break;
 	case PROP_ASYNC_CONTEXT:
 		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
 		break;
@@ -985,6 +1034,12 @@ soup_socket_start_proxy_ssl (SoupSocket *sock, const char *ssl_host,
 	if (!ssl_chan)
 		return FALSE;
 
+	/* This is optimistic, we will set this to false if we get a
+	 * cert error from one of the I/O calls
+	 */
+	if (priv->ssl_creds)
+		priv->trusted_certificate = TRUE;
+
 	priv->iochannel = ssl_chan;
 	g_io_channel_unref (real_chan);
 
@@ -1212,11 +1267,20 @@ read_from_network (SoupSocket *sock, gpointer buffer, gsize len,
 		return SOUP_SOCKET_ERROR;
 	}
 
+again:
 	status = g_io_channel_read_chars (priv->iochannel,
 					  buffer, len, nread, &my_err);
 	if (my_err) {
-		if (my_err->domain == SOUP_SSL_ERROR &&
-		    my_err->code == SOUP_SSL_ERROR_HANDSHAKE_NEEDS_WRITE)
+		if (g_error_matches (my_err, SOUP_SSL_ERROR,
+				     SOUP_SSL_ERROR_CERTIFICATE) &&
+		    !priv->ssl_strict) {
+			priv->trusted_certificate = FALSE;
+			g_clear_error (&my_err);
+			goto again;
+		}
+
+		if (g_error_matches (my_err, SOUP_SSL_ERROR,
+				     SOUP_SSL_ERROR_HANDSHAKE_NEEDS_WRITE))
 			cond = G_IO_OUT;
 		g_propagate_error (error, my_err);
 	}
@@ -1511,11 +1575,20 @@ soup_socket_write (SoupSocket *sock, gconstpointer buffer,
 		return SOUP_SOCKET_WOULD_BLOCK;
 	}
 
+again:
 	status = g_io_channel_write_chars (priv->iochannel,
 					   buffer, len, nwrote, &my_err);
 	if (my_err) {
-		if (my_err->domain == SOUP_SSL_ERROR &&
-		    my_err->code == SOUP_SSL_ERROR_HANDSHAKE_NEEDS_READ)
+		if (g_error_matches (my_err, SOUP_SSL_ERROR,
+				     SOUP_SSL_ERROR_CERTIFICATE) &&
+		    !priv->ssl_strict) {
+			priv->trusted_certificate = FALSE;
+			g_clear_error (&my_err);
+			goto again;
+		}
+
+		if (g_error_matches (my_err, SOUP_SSL_ERROR,
+				     SOUP_SSL_ERROR_HANDSHAKE_NEEDS_READ))
 			cond = G_IO_IN;
 		g_propagate_error (error, my_err);
 	}
